/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.persistence.lib;

import org.objectweb.perseus.persistence.api.ConnectionHolder;
import org.objectweb.perseus.persistence.api.TransactionalWorkingSet;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.fractal.api.control.BindingController;

import java.util.Set;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * It is an implementation of the WorkingSet interface. This implementation is
 * a Fractal component which needs a MonologFactory.
 * @author S.Chassande-Barrioz, Y.Bersihand
 */
public class BasicWorkingSet
		implements TransactionalWorkingSet, BindingController {

	protected Logger logger = null;
	protected LinkedHashMap oid2state;
	protected byte status;
	protected Object userObject;
	protected ConnectionHolder connectionHolder;
	protected boolean retainValues = true;
	protected boolean restoreValues = false;

	public BasicWorkingSet() {
		status = CTX_ACTIVE;
		oid2state = new LinkedHashMap();
		userObject = null;
		connectionHolder = null;
	}

	// IMPLEMENTATION OF THE UserBindingController INTERFACE //
	//-------------------------------------------------------//

    public String[] listFc() {
        return new String[] { };
    }

	public Object lookupFc(String c) {
		return null;
	}

	public void bindFc(String c, Object s) {
		if ("logger".equals(c)) {
			logger = (Logger) s;
		}
	}

	public void unbindFc(String c) {
	}

	// IMPLEMENTATION OF THE WorkingSet INTERFACE //
	//--------------------------------------------//

	public byte getStatus() {
		return status;
	}

	public void setStatus(byte status) throws PersistenceException {
		this.status = status;
	}

	/**
	 * @return the entry attached to the transaction weither its identifier
	 */
	public State lookup(Object oid) {
		return (State) oid2state.get(oid);
	}

	/**
	 * Attaches an entry to the transaction.
	 * @param state the state which must be attached to the transaction
	 * @param mode see WorkingSet.READ_INTENTION WorkingSet.WRITE_INTENTION WorkingSet.UNKNOWN_INTENTION
	 * @return the old state value in the cache
	 */
	public State bind(State state, Object oid, byte mode) {
		Object old;
		synchronized(oid2state) {
			old = oid2state.put(oid, state);
		}
		if (logger.isLoggable(BasicLevel.DEBUG)) {
			if (old == null) {
				logger.log(BasicLevel.DEBUG,
						"Bind (" + oid + ", " + state + ") to the context");
			} else {
				logger.log(BasicLevel.DEBUG,
						"Change the state in the context"
						+ "\n-oid=" + oid
						+ "\n-old state=" + old
						+ "\n- new state=" + state);
			}
		}
		return (State)old;
	}

	/**
	 * Dettaches an entry from the transaction.
	 * @param oid is the identifier of entry which must be detattached from the transaction
	 */
	public boolean unbind(Object oid) {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "UnBind " + oid + " from the context");
		synchronized(oid2state) {
			return (oid2state.remove(oid) == null);
		}
	}

	/**
	 * Removes all entries of the transaction.
	 */
	public void clear() {
		synchronized(oid2state) {
        	oid2state.clear();
		}
	}

	/**
	 * @return the collection of entry attached to the transaction.
	 */
	public Set entries() {
		synchronized(oid2state) {
			return new HashSet(oid2state.values());
		}
	}

    public Set oids() {
        return oid2state.keySet();
    }

	public Object getUserObject() {
		return userObject;
	}

	public ConnectionHolder getConnectionHolder() {
		return connectionHolder;
	}

	public boolean getWSRetainValues() {
		return retainValues;
	}

	public void setWSRetainValues(boolean val) {
		retainValues = val;
	}

	public boolean getWSRestoreValues() {
		return restoreValues;
	}

	public void setWSRestoreValues(boolean val) {
		restoreValues = val;
	}
}
