/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.persistence.lib;

import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.perseus.cache.api.CacheFullException;
import org.objectweb.perseus.cache.api.CacheManager;
import org.objectweb.perseus.cache.api.FixableCacheEntry;
import org.objectweb.perseus.cache.api.UnbindManager;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.api.ConcurrencyManager;
import org.objectweb.perseus.concurrency.api.RolledBackConcurrencyException;
import org.objectweb.perseus.persistence.api.MemoryInstanceManager;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.RolledBackPersistenceException;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.perseus.persistence.api.StateFilter;
import org.objectweb.perseus.persistence.api.StateManager;
import org.objectweb.perseus.persistence.api.StorageManager;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.persistence.api.TransactionalWorkingSet;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.perseus.persistence.api.WorkingSetLifeCycle;
import org.objectweb.perseus.persistence.api.WorkingSetManager;
import org.objectweb.perseus.persistence.api.VirtualState;
import org.objectweb.perseus.persistence.api.PersistentObjectLifeCycleExcpetion;
import org.objectweb.perseus.persistence.api.NoDSIPersistenceException;
import org.objectweb.perseus.persistence.concurrency.NoDSIConcurrencyException;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.Iterator;
import java.util.Set;
import java.util.ArrayList;

/**
 * This class is an implementation of the TransactionalPersistenceManager
 * interface. It manages the persistency, the concurrency and the caching of
 * persistent objects. To do this it uses several other components:
 * - a cache manager,
 * - a concurrency manager,
 * - a storage manager,
 * - a state manager,
 * - a memory instance manager,
 * - a working set manager.
 * This implementation is compliant to fractal by implementing the
 * BindingController interface permitting the management of bindings between
 * components.
 *
 * @author S.Chassande-Barrioz, P.Dechamboux, O.Lobry, Y. Bersihand
 */
public class TransactionalPersistenceManagerImpl
		implements TransactionalPersistenceManager, BindingController {

	public final static String CACHE_MANAGER_BINDING = "cache-manager";
	public final static String CONCURRENCY_MANAGER_BINDING = "concurrency-manager";
	public final static String STORAGE_MANAGER_BINDING = "storage-manager";
	public final static String STATE_MANAGER_BINDING = "state-manager";
	public final static String MEMORY_INSTANCE_MANAGER_BINDING = "memory-instance-manager";
	public final static String UNBIND_MANAGER_BINDING = "unbind-manager";
	public final static String WORKINGSET_MANAGER_BINDING = "workingset-manager";

	/**
	 * The cache of the persistent object.
	 */
	protected CacheManager cache;

	protected UnbindManager unbindManager;

	/**
	 * The concurrency manager
	 */
	protected ConcurrencyManager cm;

	/**
	 * The storage manager in charge of the I/O to the data support, and the
	 * naming management.
	 */
	protected StorageManager storage;

	/**
	 * The manager of state of persistent object.
	 */
	protected StateManager stateManager;

	/**
	 * The manager of working set instance.
	 */
	protected WorkingSetManager wsManager;

	/**
	 * The factory of persistent objects
	 */
	protected MemoryInstanceManager mim;

	protected Logger logger = null;

	/**
	 * @deprecated 
	 * Attaches a state to a working Set. 
	 *
	 * @param state is the state of the persistent object which must be attached
	 * to working set.
	 * @param ws is the working set to attach the state.
	 *
	 * @throws RolledBackPersistenceException if the entry has been evicted
	 * from the cache during the locking step (concurrency manager). It is
	 * possible that it has been removed by another thread ==> rollback the
	 * working set.
	 */
	protected void bindCeInWS(State state, WorkingSet ws, byte mode)
			throws PersistenceException {
		Object oid = state.getCacheEntry().getCeIdentifier();
		ws.bind(state, oid, WorkingSet.UNKNOWN_INTENTION);
	}


	/**
	 * finds or creates the CacheEntry matching to an identifier. If no cache
	 * antry matching the identifier is found in the cache, an new instance
	 * is created without state.
	 * @param ws if the WorkingSet requiring a CacheEntry.
	 * @param oid the identifier of the expected cache entry
	 * @return a cache entry instance (never null).
	 * @throws PersistenceException if it is not possible to instanciate a
	 * persistent object
	 * @throws RolledBackPersistenceException if the cache is full (case of the
	 * entry has not been found into the cache and a new entry is bound).
	 */
	protected CacheEntry getCacheEntry(WorkingSet ws,
									   Object oid) throws PersistenceException {
		CacheEntry ce = cache.lookup(oid);
		if (ce == null) {
			// The element is not in the cache => load it and put it in the
			// working set.
			Object obj = null;
			// create memory instance if lookup failed
			obj = mim.newInstance(oid, ws.getConnectionHolder());
			// add memory instance to cache
			try {
				ce = cache.bind(oid, obj);
				//Remove the reference in order to force the data loading
				stateManager.setReferenceState(ce, null);
			} catch (CacheFullException e) {
				throw new RolledBackPersistenceException(e);
			} catch (CacheException e) {
				//That means the entry has been already bind to the cache. There
				// are concurrent accesses.
				ce = cache.lookup(oid);
			}
		}
		return ce;
	}

	/**
	 * Check if a cache entry is already referenced in the cache. If the entry
	 * has not been found, the entry is bound in the cache.
	 *
	 * @param ce is the cache entry to check
	 * @throws PersistenceException if the cache is full or if another cache
	 * entry has been found in the cache with the same identifier.
	 */
	protected void checkCacheEntry(CacheEntry ce) throws PersistenceException {
		Object oid = ce.getCeIdentifier();
		if (((FixableCacheEntry) ce).getCeFixCount() == 0) {
			CacheEntry _ce = cache.lookup(oid);
			if (_ce == null) {
				//the entry is not present in the cache
				try {
					ce = cache.bind(oid, ce.getCeObject());
				} catch (CacheFullException e) {
					throw new RolledBackPersistenceException(e);
				} catch (CacheException e) {
					//That means the entry has been already bind to the cache. There
					// are concurrent accesses.
					ce = cache.lookup(oid);
				}
			} else if (ce != _ce) {
				//the entry present in the cache is not the same than the parameter
				throw new RolledBackPersistenceException(
						"Another persistent object is already present into the cache (identifier=" + oid + ")");
			}
		}
	}

	/**
	 * Prepares the end of a working set.
	 * @param ws is can be simple working set or a transaction
	 * @return true if the context CAN be committed, false if the context MUST be
	 * rolled back
	 * @throws PersistenceException
	 */
	private boolean prepareWS(WorkingSet ws, boolean validate)
			throws PersistenceException {
		boolean valid = validate;
		PersistenceException pe = null;
		try {
			ws.setStatus(WorkingSetLifeCycle.getNextStatus(
					ws.getStatus(), WorkingSetLifeCycle.PREPARE_ACTION));
		} catch (PersistenceException e) {
			pe = e;
			valid = false;
		}
		try {
			// check if the context can be validate on concurrency manager
			valid = cm.validate(ws) && valid;
			if (pe != null) {
				throw pe;
			}
			if (valid) {
				//flush dirty objects into data support,
				Set entries = ws.entries();
				for (Iterator it = entries.iterator(); it.hasNext();) {
					State state = (State) it.next();
					if (state == VirtualState.instance) {
						continue;
					}
					if (stateManager.isDirty(state)) {
						storage.write(ws.getConnectionHolder(),
								state.getCacheEntry().getCeIdentifier(),
								state);
					}
				}
			}
		} catch(Exception e) {
			// If an exception occurs then the working set must be cancelled
			valid = false;
			throw (e instanceof PersistenceException
				? (PersistenceException) e
				: new PersistenceException("An exception occurs during the "
					+ "working preparation, it is marked as rolled back",e));
		} finally {
			//release the data support connection
			ws.getConnectionHolder().releaseCHConnection();
			if (valid) {
				ws.setStatus(WorkingSetLifeCycle.getNextStatus(
						ws.getStatus(), WorkingSetLifeCycle.PREPARE_OK_ACTION));
			} else {
				ws.setStatus(WorkingSetLifeCycle.getNextStatus(
						ws.getStatus(), WorkingSetLifeCycle.PREPARE_FAIL_ACTION));
			}
		}
		return valid;
	}

	/**
	 * commit a context.
	 * @param ws is can be simple working set or a transaction
	 * @param action is the action declenching the rollback
	 * @throws PersistenceException
	 */
	private void commitWS(WorkingSet ws,
						  byte action) throws PersistenceException {
		byte newStatus = WorkingSetLifeCycle.getNextStatus(
				ws.getStatus(), action);
		try {
			ws.getConnectionHolder().commitCH();
		} catch (PersistenceException e) {
			rollbackWS(ws, WorkingSetLifeCycle.ROLLBACK_ACTION);
			throw new RolledBackPersistenceException(e);
		}

		try {
			Set entries = ws.entries();
			ArrayList unexportedOids = null;
			// copies accessor states into proxies,
			for (Iterator it = entries.iterator(); it.hasNext();) {
				State state = (State) it.next();
				if (state == VirtualState.instance) {
					continue;
				}
				CacheEntry ce = state.getCacheEntry();
				if (stateManager.isUnexported(state)) {
					if (unexportedOids == null) {
						unexportedOids = new ArrayList();
					}
					unexportedOids.add(ce.getCeIdentifier());
					stateManager.makeUnbound(ce);
				} else {
					if (ws.getWSRetainValues()) {
						stateManager.makeClean(state);
						stateManager.setReferenceState(ce, state);
					} else if (stateManager.getReferenceState(ce) == state) {
						stateManager.setReferenceState(ce, null);
					} else {
						stateManager.destroyState(state);
					}
				}
			}
			if (unexportedOids != null) {
				// unbind removed entries from the cache
				unbindManager.unbindAll(unexportedOids, true);
			}
		} catch (CacheException e) {
			throw new PersistenceException(e);
		} finally {
			try {
				ws.getConnectionHolder().closeCHConnection();
			} finally {
				cm.finalize(ws);
				ws.clear();
				ws.setStatus(newStatus);
				storage.endWS(ws);
			}
		}
	}

	/**
	 * cancel a context. The unfix action is out of date.
	 * @param ws is can be simple working set or a transaction
	 * @param action is the action declenching the rollback
	 * @throws PersistenceException
	 */
	private void rollbackWS(WorkingSet ws,
							byte action) throws PersistenceException {
		byte newStatus = WorkingSetLifeCycle.getNextStatus(
				ws.getStatus(), action);
		ArrayList errors = null;
		try {
			try {
				ws.getConnectionHolder().rollbackCH();
			} finally {
				ws.setStatus(newStatus);
			}
			ArrayList unexportedOids = null;
			for (Iterator i = ws.entries().iterator(); i.hasNext();) {
				State state = (State) i.next();
				if (state == VirtualState.instance) {
					continue;
				}
				CacheEntry ce = state.getCacheEntry();
				if (stateManager.isExported(state)) {
					if (unexportedOids == null) {
						unexportedOids = new ArrayList();
					}
					unexportedOids.add(ce.getCeIdentifier());
					stateManager.makeUnbound(ce);
				}
				if (stateManager.isDirty(state)) {
					if (state == stateManager.getReferenceState(ce)) {
						//The reference state
						if (ws.getWSRestoreValues()) {
							//Restore value from the support
							try {
								storage.read(ws.getConnectionHolder(), ce.getCeIdentifier(), state);
							} catch (Exception e) {
								//catch erros in order to finalize the cache eviction
								if (errors == null) {
									errors = new ArrayList(5);
								}
								errors.add(e);
							}
						} else {
							//At the next use it will be loaded
							stateManager.setReferenceState(ce, null);
							stateManager.destroyState(state);
						}
					} else {
						stateManager.destroyState(state);
					}
				} else if (state != stateManager.getReferenceState(ce)) {
					stateManager.destroyState(state);
				} //else reference state is already clean
			}
			if (unexportedOids != null) {
				// unbind removed entries from the cache
				unbindManager.unbindAll(unexportedOids, true);
			}
			if (errors != null) {
				Exception e = (Exception) errors.get(0);
				if (e instanceof PersistenceException) {
					throw (PersistenceException) e;
				} else {
					throw new PersistenceException(e);
				}
			}
		} catch (CacheException ce) {
			throw new PersistenceException("Impossible to rollback - ", ce);
		} finally {
			try {
				ws.getConnectionHolder().closeCHConnection();
			} finally {
				cm.abort(ws);
				ws.clear();
				ws.setStatus(newStatus);
				storage.endWS(ws);
			}
		}
	}

	private void assertOpenWorkingset(WorkingSet ws) throws PersistenceException {
		if (ws.getStatus() == WorkingSet.CTX_CLOSED) {
			throw new PersistenceException("Impossible to work with a closed WorkingSet");
		}
	}

	//IMPLEMENTATION OF THE UserBindingControler INTERFACE //
	//------------------------------------------------------//

	public String[] listFc() {
		return new String[]{
			CACHE_MANAGER_BINDING,
			CONCURRENCY_MANAGER_BINDING,
			STATE_MANAGER_BINDING,
			STORAGE_MANAGER_BINDING,
			MEMORY_INSTANCE_MANAGER_BINDING,
			UNBIND_MANAGER_BINDING,
			WORKINGSET_MANAGER_BINDING,
			STATE_MANAGER_BINDING
		};
	}

	public Object lookupFc(String s) {
		if (CACHE_MANAGER_BINDING.equals(s)) {
			return cache;
		} else if (CONCURRENCY_MANAGER_BINDING.equals(s)) {
			return cm;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			return stateManager;
		} else if (STORAGE_MANAGER_BINDING.equals(s)) {
			return storage;
		} else if (MEMORY_INSTANCE_MANAGER_BINDING.equals(s)) {
			return mim;
		} else if (UNBIND_MANAGER_BINDING.equals(s)) {
			return unbindManager;
		} else if (WORKINGSET_MANAGER_BINDING.equals(s)) {
			return wsManager;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			return stateManager;
		}
		return null;
	}

	public void bindFc(String s, Object o) {
		if ("logger".equals(s)) {
			logger = (Logger) o;
		} else if (CACHE_MANAGER_BINDING.equals(s)) {
			cache = (CacheManager) o;
		} else if (CONCURRENCY_MANAGER_BINDING.equals(s)) {
			cm = (ConcurrencyManager) o;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			stateManager = (StateManager) o;
		} else if (STORAGE_MANAGER_BINDING.equals(s)) {
			storage = (StorageManager) o;
		} else if (MEMORY_INSTANCE_MANAGER_BINDING.equals(s)) {
			mim = (MemoryInstanceManager) o;
		} else if (UNBIND_MANAGER_BINDING.equals(s)) {
			unbindManager = (UnbindManager) o;
		} else if (WORKINGSET_MANAGER_BINDING.equals(s)) {
			wsManager = (WorkingSetManager) o;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			stateManager = (StateManager) o;
		}
	}

	public void unbindFc(String s) {
		if (CACHE_MANAGER_BINDING.equals(s)) {
			cache = null;
		} else if (CONCURRENCY_MANAGER_BINDING.equals(s)) {
			cm = null;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			stateManager = null;
		} else if (STORAGE_MANAGER_BINDING.equals(s)) {
			storage = null;
		} else if (MEMORY_INSTANCE_MANAGER_BINDING.equals(s)) {
			mim = null;
		} else if (UNBIND_MANAGER_BINDING.equals(s)) {
			unbindManager = null;
		} else if (WORKINGSET_MANAGER_BINDING.equals(s)) {
			wsManager = null;
		} else if (STATE_MANAGER_BINDING.equals(s)) {
			stateManager = null;
		}
	}

	// IMPLEMENTATION OF THE PersistenceManager INTERFACE //
	//----------------------------------------------------//

	public State export(WorkingSet ws, Object obj)
			throws PersistenceException {
		return export(ws, obj, null);
	}

	public State export(WorkingSet ws, Object obj, Object hints)
			throws PersistenceException {
		if (logger.isLoggable(BasicLevel.DEBUG))
			logger.log(BasicLevel.DEBUG, "export ctx=" + ws
					+ " / obj=" + obj.getClass().getName()
					+ " / hints=" + hints);
		assertOpenWorkingset(ws);
		State state = null;
		//export through storage manager
		Object oid = null;
		try {

			if (hints == null) {
				oid = storage.export(ws.getConnectionHolder(), obj);
			} else {
				oid = storage.export(ws.getConnectionHolder(), obj, hints);
			}

			//add to cache
			CacheEntry ce = cache.bind(oid, obj);

			//notify concurrency manager
			state = writeIntention(ws, ce);
			stateManager.makeExported(state);
			return state;
		} catch (PersistenceException e) {
			throw e;
		} catch (CacheFullException e) {
			throw new RolledBackPersistenceException(e);
		} catch (CacheException e) {
			throw new PersistenceException("Cannot put object into cache", e);
		}
	}

	public State unexport(WorkingSet ws, Object oid) throws PersistenceException {
		return unexport(ws, getCacheEntry(ws, oid));
	}

	public State unexport(WorkingSet ws, CacheEntry ce)
			throws PersistenceException {
		assertOpenWorkingset(ws);
		if (!stateManager.isBound(ce)) {
			throw new PersistentObjectLifeCycleExcpetion(
					PersistentObjectLifeCycleExcpetion.UNEXPORT_NPO_MSG);
		}
		Object oid = ce.getCeIdentifier();
		State state = null;
		//notify concurrency manager
		state = writeIntention(ws, ce);
		//unexport through storage manager
		storage.unexport(ws.getConnectionHolder(), oid, ce);
		stateManager.makeUnexported(state);
		return state;
	}

	public State readIntention(WorkingSet ws, Object oid)
			throws PersistenceException {
		return readIntention(ws, getCacheEntry(ws, oid));
	}

	/**
	 * This method records an access intention to a data object in read mode.
	 * @param ce is the cache entry
	 * @param ws is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @exception PersistenceException
	 */
	public State readIntention(WorkingSet ws, CacheEntry ce) throws PersistenceException {
		assertOpenWorkingset(ws);
		if (!stateManager.isBound(ce)) {
			throw new PersistentObjectLifeCycleExcpetion(
					PersistentObjectLifeCycleExcpetion.READ_NPO_MSG);
		}
		checkCacheEntry(ce);
		State state;
		// notify concurrency manager
		try {
			state = (State) cm.readIntention(ws, ce.getCeIdentifier(), ce);
		} catch (RolledBackConcurrencyException e) {
			throw new RolledBackPersistenceException(e);
		} catch (NoDSIConcurrencyException e) {
			try {
				unbindManager.unbind(ce.getCeIdentifier(), true);
			} catch (CacheException e1) {
				logger.log(BasicLevel.ERROR,
						"Impossible to unbind from the cache an entry which " +
						"does not exist on the data support", e1);
			}
			throw new NoDSIPersistenceException(e);
		} catch (ConcurrencyException e) {
			throw new PersistenceException(e);
		}
		Object oid = state.getCacheEntry().getCeIdentifier();
		ws.bind(state, oid, WorkingSet.READ_INTENTION);
		return state;
	}

	public State writeIntention(WorkingSet ws, Object oid)
			throws PersistenceException {
		return writeIntention(ws, getCacheEntry(ws, oid));
	}

	public State writeIntention(WorkingSet ws, CacheEntry ce)
			throws PersistenceException {
		assertOpenWorkingset(ws);
		if (!stateManager.isBound(ce)) {
			throw new PersistentObjectLifeCycleExcpetion(
					PersistentObjectLifeCycleExcpetion.WRITE_NPO_MSG);
		}
		checkCacheEntry(ce);
		State state;
		// notify concurrency manager
		try {
			state = (State) cm.writeIntention(ws, ce.getCeIdentifier(), ce);
		} catch (RolledBackConcurrencyException e) {
			throw new RolledBackPersistenceException(e);
		} catch (NoDSIConcurrencyException e) {
			try {
				unbindManager.unbind(ce.getCeIdentifier(), true);
			} catch (CacheException e1) {
				logger.log(BasicLevel.ERROR,
						"Impossible to unbind from the cache an entry which " +
						"does not exist on the data support", e1);
			}
			throw new NoDSIPersistenceException(e);
		} catch (ConcurrencyException e) {
			throw new PersistenceException(e);
		}
		Object oid = state.getCacheEntry().getCeIdentifier();
		ws.bind(state, oid, WorkingSet.WRITE_INTENTION);
		stateManager.makeDirty(state);
		return state;
	}

	/**
	 * This method can be used to mark the end of an access to a data object.
	 * @param ws is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param ce is the  cache entry on which the access is finished.
	 * @exception PersistenceException
	 */
	public void accessCompletion(WorkingSet ws, State ce) throws PersistenceException {
		assertOpenWorkingset(ws);
	}

	/**
	 * This method permits to flush entries associated to a context.
	 * @param ws is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param statefilter is a filter of the state which must be flushed
	 * @throws PersistenceException if the context is not valid.
	 */
	public void flush(WorkingSet ws, StateFilter statefilter)
			throws PersistenceException {
		assertOpenWorkingset(ws);
		Iterator it = ws.entries().iterator();
		while (it.hasNext()) {
			State state = (State) it.next();
			if (state == VirtualState.instance) {
				continue;
			}
			if (stateManager.isDirty(state) && statefilter.accept(state)) {
				storage.write(ws.getConnectionHolder(),
						state.getCacheEntry().getCeIdentifier(),
						state);
				//Mark the entry as flushed in order to avoid I/O at prepare time
				stateManager.makeFlushed(state);
			}
		}
	}

	/**
	 * This method permits to flush entries associated to a context.
	 * @param ws is an identifier of the execution context. It can be a
	 * transaction handle.
	 * @param state is the state taht must be flushed
	 * @throws PersistenceException if the context is not valid.
	 */
	public void flush(WorkingSet ws, State stateToFlush)
			throws PersistenceException {
		assertOpenWorkingset(ws);
		Iterator it = ws.entries().iterator();
		while (it.hasNext()) {
			State state = (State) it.next();
			if (state == VirtualState.instance) {
				continue;
			}
			if (stateManager.isDirty(state) && state == stateToFlush) {
				storage.write(ws.getConnectionHolder(),
						state.getCacheEntry().getCeIdentifier(),
						state);
				//Mark the entry as flushed in order to avoid I/O at prepare time
				stateManager.makeFlushed(state);
			}
		}
	}
	/**
	 * The unfix action is out of date.
	 */
	public boolean evict(WorkingSet ws, Object oid, boolean force) throws PersistenceException {
		FixableCacheEntry ce = (FixableCacheEntry) cache.lookup(oid);
		assertOpenWorkingset(ws);
		if (ce == null) {
			return true;
		}
		int fix = ce.getCeFixCount();
		State state = ws.lookup(oid);
		if (fix == 1 && state != null) {
			//The entry is used only in the specified working set
			if (stateManager.isDirty(state)) {
				//flush the dirty object and mark it as flushed
				storage.write(ws.getConnectionHolder(), oid, state);
				stateManager.makeFlushed(state);
			}
			//remove the entry from the working set by replaceing the state
			// by a virtual state.
			ws.bind(VirtualState.instance, oid, WorkingSet.UNKNOWN_INTENTION);
		}
		if (fix > 0) {
			if (force) {
				throw new PersistenceException("Impossible to evict an entry from the cache: There is another user than the given working set");
			} else {
				return false;
			}
		}
		//Remove the reference state
		stateManager.setReferenceState(ce, null);
		//Remove the local reference for the GC
		ce = null;
		//remove the entry from the cache
		try {
			return unbindManager.unbind(oid, false);
		} catch (CacheException e) {
			if (force) {
				throw new PersistenceException("Impossible to unbind the entry: ", e);
			} else {
				return false;
			}
		}
	}

	public int evictAll(WorkingSet ws, boolean force) throws PersistenceException{
		assertOpenWorkingset(ws);
		try {
			return unbindManager.unbindUnfixed(force).size();
		} catch (CacheException e) {
			throw new PersistenceException(e);
		}
	}

	public void unbind(WorkingSet ws, Object oid) throws PersistenceException {
		assertOpenWorkingset(ws);
		unbind(ws, cache.lookup(oid));
	}

	public void unbind(WorkingSet ws, CacheEntry ce) throws PersistenceException {
		assertOpenWorkingset(ws);
		if (ce == null || stateManager.getReferenceState(ce) == null) {
			return;
		}
		FixableCacheEntry fce = (FixableCacheEntry) ce;
		int fix = fce.getCeFixCount();
		Object oid = ce.getCeIdentifier();
		State state = (ws == null ? null : ws.lookup(oid));
		if (fix == 1 && state != null) {
			//The entry is used only in the specified working set
			if (stateManager.isDirty(state)) {
				//flush the dirty object and mark it as flushed
				storage.write(ws.getConnectionHolder(), oid, state);
				stateManager.makeFlushed(state);
			}
			try {
				cache.unfix(ce);
			} catch (CacheException e) {
					throw new PersistenceException(
							"Impossible to unfix the entry: ", e);
			}
			fix = fce.getCeFixCount();

			// Remove the entry from the working set by replacing the real state
			// by a virtual state. It is required to keep the oid into the
			// working set in order to free the locking of the concurrency
			// manager
			ws.bind(VirtualState.instance, oid, WorkingSet.UNKNOWN_INTENTION);
		}
		if (fix > 0) {
				throw new PersistenceException(
						"There is another user than the given working set");
		}

		//remove the entry from the cache
		try {
			unbindManager.unbind(oid, true);
		} catch (CacheException e) {
				throw new PersistenceException("Impossible to unbind the entry: ", e);
		}

		//Mark the cache entry as unbound from the Persistence system
		stateManager.makeUnbound(ce);
		if (state != null) {
			//Use the current state as reference state
			stateManager.setReferenceState(ce, state);
		} // no state associated to the working set ==> use the reference state
	}

	public void refresh(WorkingSet ws, Object oid) throws PersistenceException {
		assertOpenWorkingset(ws);
		refresh(ws, cache.lookup(oid));
	}
	public void refresh(WorkingSet ws, CacheEntry ce) throws PersistenceException {
		assertOpenWorkingset(ws);
		State refState = stateManager.getReferenceState(ce);
		if (ce == null || refState == null) {
			return;
		}
		Object oid = ce.getCeIdentifier();
		State state = (ws == null ? null : ws.lookup(oid));
		if (stateManager.isDirty(state)) {
			if (refState == state) {
				//forget the reference state
				stateManager.setReferenceState(ce, null);
				state = VirtualState.instance;
			} else {
				state = refState;
			}
		} else { // isClean
			//re-read from the reference state from the cache or later from the
			// data support
			if (refState == null) { //forget value
				state = VirtualState.instance;
			} else { //from the cache
				state = refState;
			}
		}
		ws.bind(state, oid, WorkingSet.UNKNOWN_INTENTION);
	}

	/**
	 * creates a working set
	 * @param userObject can be an external user depending on the
	 * personnality.
	 * @return a new WorkingSet
	 * @throws PersistenceException
	 */
	public WorkingSet createWS(Object userObject)
			throws PersistenceException {
		WorkingSet ws = wsManager.createWS(userObject);
		storage.beginWS(ws);
		return ws;
	}

	/**
	 * creates a working set
	 * @param userObject can be an external user depending on the
	 * personnality.
	 * @param workingSetType is a type of the working set.
	 * @return a new WorkingSet
	 * @throws PersistenceException
	 */
	public WorkingSet createWS(Object userObject,
							   Object workingSetType)
			throws PersistenceException {
		WorkingSet ws = wsManager.createWS(userObject, workingSetType);
		storage.beginWS(ws);
		return ws;
	}

	/**
	 * Closes and releases a context.
	 * @param context which must be closed
	 * @throws PersistenceException if the context is a transactional context
	 * which has not been commited or aborted.
	 */
	public void close(WorkingSet context) throws PersistenceException {
		if (!context.entries().isEmpty()) {
			boolean prepared_ok = false;
			PersistenceException pe = null;
			try {
				prepared_ok = prepareWS(context, true);
			} catch (PersistenceException e) {
				pe = e;
			}
			if (prepared_ok) {
				commitWS(context, WorkingSetLifeCycle.CLOSE_ACTION);
			} else {
				rollbackWS(context, WorkingSetLifeCycle.CLOSE_ACTION);
				if (pe == null) {
					throw new RolledBackPersistenceException();
				} else {
					throw pe;
				}
			}
		} else {
			context.getConnectionHolder().closeCHConnection();
			context.setStatus(WorkingSetLifeCycle.getNextStatus(
					context.getStatus(), WorkingSetLifeCycle.CLOSE_ACTION));
			storage.endWS(context);
		}
		wsManager.closeWS(context);
	}

	// IMPLEMENTATION OF THE TransactionalPersistenceManager INTERFACE //
	//-----------------------------------------------------------------//

	/**
	 * converts a context to a transaction
	 * @param tx becomes transactional
	 * @throws PersistenceException if the context is already transactional.
	 */
	public void begin(TransactionalWorkingSet tx) throws PersistenceException {
		assertOpenWorkingset(tx);
		// 1st step: flush objects in working set, outside transaction
		if (tx.entries().size() > 0) {
			logger.log(BasicLevel.DEBUG, "Flush modification before the transaction " + tx);
			close(tx);
		}
		tx.getConnectionHolder().begin();

		// 2nd step: change the transaction's status
		tx.setStatus(WorkingSetLifeCycle.getNextStatus(
				tx.getStatus(), WorkingSetLifeCycle.BEGIN_TX_ACTION));
	}

	/**
	 * Prepare the transactinal context.
	 * @param tx the
	 * @return true if the transaction can be commit or false is the transaction
	 * must be rolled back.
	 * @throws PersistenceException if an error raised during the prepare
	 * operation, or if the context is not a transactional context.
	 */
	public boolean prepare(TransactionalWorkingSet tx) throws PersistenceException {
		assertOpenWorkingset(tx);
		return prepareWS(tx, true);
	}

	/**
	 * commits a transactional context. The call to the prepare method is
	 * required.
	 * @param tx
	 * @throws PersistenceException if an error raised during the commit
	 * operation, or if the context is not a transactional context.
	 */
	public void commit(TransactionalWorkingSet tx) throws PersistenceException {
		assertOpenWorkingset(tx);
		if (tx.getStatus() == TransactionalWorkingSet.CTX_ACTIVE_TRANSACTIONAL) {
			//One phase commit
			boolean prepared_ok = false;
			PersistenceException pe = null;
			try {
				prepared_ok = prepareWS(tx, true);
			} catch (PersistenceException e) {
				pe = e;
			}
			if (prepared_ok) {
				commitWS(tx, WorkingSetLifeCycle.COMMIT_ACTION);
			} else {
				rollbackWS(tx, WorkingSetLifeCycle.ROLLBACK_ACTION);
				if (pe == null) {
					throw new RolledBackPersistenceException();
				} else {
					throw pe;
				}
			}
		} else {
			commitWS(tx, WorkingSetLifeCycle.COMMIT_ACTION);
		}
		storage.beginWS(tx);
	}

	/**
	 * rollbacks a transactional context. The call to the prepare method is
	 * required.
	 * @param tx
	 * @throws PersistenceException if an error raised during the rollback
	 * operation, or if the context is not a transactional context.
	 */
	public void rollback(TransactionalWorkingSet tx) throws PersistenceException {
		assertOpenWorkingset(tx);
		try {
			prepareWS(tx, false);
		} catch (PersistenceException e) {
			logger.log(BasicLevel.WARN, "An error occris during the rollback operation", e);
		}
		rollbackWS(tx, WorkingSetLifeCycle.ROLLBACK_ACTION);
		storage.beginWS(tx);
	}
}
